from dataclasses import dataclass
from typing import Dict, Type

from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.dag.manager import DAGRunConcurrentManager
from ml_pipeline_engine.dag.retrying import DagRetryPolicy
from ml_pipeline_engine.parallelism import (
    process_pool_registry,
    threads_pool_registry,
)
from ml_pipeline_engine.types import (
    DAGLike,
    DAGRunManagerLike,
    NodeId,
    NodeResultT,
    NodeSerializerLike,
    PipelineContextLike,
    RetryPolicyLike,
)


@dataclass()
class DAG(DAGLike):
    graph: DiGraph
    input_node: NodeId
    output_node: NodeId
    is_process_pool_needed: bool
    is_thread_pool_needed: bool
    node_map: Dict[NodeId, NodeSerializerLike]
    retry_policy: Type[RetryPolicyLike] = DagRetryPolicy
    run_manager: Type[DAGRunManagerLike] = DAGRunConcurrentManager

    def __post_init__(self) -> None:
        self._start_runtime_validation()

    def _start_runtime_validation(self) -> None:
        self._validate_pool_executors()

    def _validate_pool_executors(self) -> None:
        if self.is_thread_pool_needed:
            threads_pool_registry.is_ready()

        if self.is_process_pool_needed:
            process_pool_registry.is_ready()

    async def run(self, ctx: PipelineContextLike) -> NodeResultT:
        return await self.run_manager(dag=self).run(ctx)



import enum


class SerializationNodeKind(str, enum.Enum):
    """
    Поддерживаемые типы сериализации узлов между
    """

    cloudpickle = 'cloudpickle'


class NodeTag(str, enum.Enum):
    """
    Поддерживаемые теги узлов
    """

    process = 'process'
    thread = 'thread'
    non_async = 'non_async'



from enum import Enum


class NodeField(str, Enum):
    is_switch = 'is_switch'
    is_first_success_pool = 'is_first_success_pool'
    is_first_success = 'is_first_success'
    start_node = 'start_node'
    max_iterations = 'max_iterations'
    additional_data = 'additional_data'


class EdgeField(str, Enum):
    kwarg_name = 'kwarg_name'
    is_switch = 'is_switch'
    case_branch = 'case_branch'
    is_first_success = 'is_first_success'



import copy
import typing as t

from ml_pipeline_engine.types import DAGCacheManagerLike, NodeId


class Cache(DAGCacheManagerLike):
    def __init__(self):
        self._cache = {}

    def save(self, node_id: NodeId, data: t.Any):
        self._cache[node_id] = copy.deepcopy(data)

    def load(self, node_id: NodeId) -> t.Any:
        if node_id in self._cache:
            return self._cache[node_id]

    def exists(self, node_id: NodeId) -> bool:
        return bool(node_id in self._cache)

    def remove(self, node_id: NodeId) -> None:
        self._cache.pop(node_id, None)



from typing import Optional

import networkx as nx
from pyvis.network import Network


class DiGraph(nx.DiGraph):

    def __init__(self, is_recurrent: bool = False, **kwargs):
        super().__init__(**kwargs)

        self.is_recurrent = is_recurrent

    def create_interactive_graph(self, height: Optional[str] = None, path: Optional[str] = None) -> None:
        """
        Creating an interactive graph via pyvis.
        As a result, you will receive an HTML file.
        """

        net = Network(
            height=height or '800px',
            directed=True,
            select_menu=True,
            filter_menu=True,
            cdn_resources='remote',
        )
        net.toggle_physics(False)
        net.show_buttons()
        net.from_nx(self)
        net.show(path or 'interactive_graph.html', notebook=False)



from enum import Enum
from typing import Optional


class NodeErrorType(str, Enum):
    succession_node_error = 'succession_node_error'


class DataSourceCollectError(Exception):
    def __init__(self, source_title: str, source_name: Optional[str] = None, *args):
        super().__init__(args)
        self.source_title = source_title
        self.source_name = source_name

    def __str__(self):
        return f'Не удалось получить данные из источника "{self.source_title}"'



import functools
import logging

from ml_pipeline_engine.base_nodes.datasources import DataSource
from ml_pipeline_engine.exceptions import DataSourceCollectError

logger = logging.getLogger(__name__)


def guard_datasource_error(name=None, title=None):
    def _guard_datasource_error(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            is_empty_names = name is None and title is None
            if args and isinstance(args[0], DataSource):
                assert is_empty_names, 'Не нужно явно указывать имя источника данных, если используется DataSource'
                data_source = args[0]
            else:
                assert not is_empty_names, 'Укажите имя источника данных явно'
                _name, _title = name, title

                class _LegacyDataSource(DataSource):  # noqa
                    """
                    Для совместимости с кодом, где не используется класс DataSource
                    """

                    name = _name
                    title = _title

                    def collect(self):
                        raise NotImplementedError

                data_source = _LegacyDataSource()
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                logger.info('Источник отработал с ошибкой и вернул DataSourceCollectError: %s', str(ex))
                return DataSourceCollectError(
                    source_title=data_source.title,
                    source_name=getattr(data_source, 'name', None),
                )

        return wrapper

    return _guard_datasource_error



import abc
import logging
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from typing import Optional, Union

logger = logging.getLogger(__name__)


class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


PoolExecutorT = Union[ProcessPoolExecutor, ThreadPoolExecutor]


class PoolExecutorRegistry(metaclass=SingletonMeta):

    def __init__(self):
        self._pool_executor: Optional[PoolExecutorT] = None

    @abc.abstractmethod
    def is_ready(self) -> None:
        ...

    def register_pool_executor(self, pool_executor: PoolExecutorT) -> None:
        if self._pool_executor:
            logger.info(
                'Регистрация пула %s возможна только один раз. Повторная инициализация будет пропущена',
                type(pool_executor),
            )
            return

        logger.info('Пул %s зарегистрирован', type(pool_executor))
        self._pool_executor = pool_executor

    def get_pool_executor(self) -> PoolExecutorT:
        self.is_ready()
        return self._pool_executor

    def shutdown(self) -> None:
        if self._pool_executor is not None:
            self._pool_executor.shutdown()

    def __del__(self) -> None:
        self.shutdown()



import typing as t

from ml_pipeline_engine.artifact_store.store.base import ArtifactStore
from ml_pipeline_engine.types import NodeId


class NoOpArtifactStore(ArtifactStore):
    async def save(self, node_id: NodeId, data: t.Any) -> None:
        ...

    async def load(self, node_id: NodeId) -> t.Any:
        raise NotImplementedError('Метод "load" не реализован для данного типа хранилища')



from ml_pipeline_engine.types import NodeBase


class MLModelBase(NodeBase):
    """
    Базовый класс для ML-моделей
    """

    node_type = 'ml_model'

    def predict(self, *args, **kwargs):
        raise NotImplementedError('Method predict() is not implemented')



import asyncio
import logging
import typing as t
from collections import deque
from contextlib import suppress
from dataclasses import dataclass

import networkx as nx

from ml_pipeline_engine.dag.enums import EdgeField, NodeField
from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.dag.retrying import DagRetryPolicy
from ml_pipeline_engine.dag.utils import get_connected_subgraph
from ml_pipeline_engine.exceptions import NodeErrorType
from ml_pipeline_engine.node import run_node, run_node_default
from ml_pipeline_engine.types import (
    CaseResult,
    DAGLike,
    DAGRunManagerLike,
    NodeId,
    NodeResultT,
    NodeSerializerLike,
    PipelineContextLike,
    Recurrent,
)

logger = logging.getLogger(__name__)


@dataclass
class DAGRunConcurrentManager(DAGRunManagerLike):
    """
    Дефолтный менеджер запуска графов.
    Производит конкурентное исполнение узлов.
    """

    dag: DAGLike

    async def run(self, ctx: PipelineContextLike) -> NodeResultT:
        return await self._run_dag(ctx, self._get_reduced_dag(self.dag.input_node, self.dag.output_node))

    async def _get_node_kwargs(self, ctx: PipelineContextLike, node_id: NodeId) -> t.Dict[str, t.Any]:
        """
        Получить входные kwarg-и для узла графа перед его запуском
        """

        kwargs = {}

        if node_id != self.dag.input_node:
            for pred_node_id in self.dag.graph.predecessors(node_id):
                kwarg_name = self.dag.graph.edges[(pred_node_id, node_id)].get(EdgeField.kwarg_name)

                if kwarg_name is None:
                    continue

                if self.dag.graph.nodes[pred_node_id].get(NodeField.is_switch):
                    kwargs[kwarg_name] = await ctx.load_node_result(
                        (await ctx.get_case_result(pred_node_id)).node_id
                    )
                    continue

                kwargs[kwarg_name] = await ctx.load_node_result(pred_node_id)

        else:
            kwargs = ctx.input_kwargs

        additional_data = self.dag.graph.nodes[node_id].get(NodeField.additional_data)

        if additional_data:
            kwargs[NodeField.additional_data] = additional_data

        return kwargs

    @staticmethod
    def _is_switch(dag: DiGraph, node_id: NodeId) -> bool:
        """
        Является ли узел switch-узлом
        """

        try:
            return dag.nodes[node_id].get(NodeField.is_switch) is True
        except KeyError:
            return False

    @staticmethod
    def _is_node_in_oneof(dag: DiGraph, node_id: NodeId) -> bool:
        """
        Является ли узел узлом InputOneOf пула
        """
        return bool(dag.nodes[node_id].get(NodeField.is_first_success_pool))

    @staticmethod
    def _is_head_of_oneof(dag: DiGraph, node_id: NodeId) -> bool:
        """
        Является ли узел родителем InputOneOf пула
        """
        return bool(dag.nodes[node_id].get(NodeField.is_first_success))

    def _get_reduced_dag(self, source: NodeId, dest: NodeId) -> DiGraph:
        """
        Получить связный подграф между двумя заданными узлами графа с удаленными ребрами условных операторов
        """

        def _filter(u, v) -> bool:
            """
            Удаляет ребра с меткой EdgeField.case_branch из subgraph_view

            Args:
                u - Node
                v - Node Edge
            """
            return not self.dag.graph.edges[u, v].get(EdgeField.case_branch)

        def _filter_node(u) -> bool:
            """
            Удаляет ноды с меткой NodeField.is_first_success_pool из subgraph_view

            Args:
                u -  Node
            """
            return not self.dag.graph.nodes[u].get(NodeField.is_first_success_pool)

        return get_connected_subgraph(
            nx.subgraph_view(self.dag.graph, filter_edge=_filter, filter_node=_filter_node), source, dest
        )

    def _get_reduced_dag_input_one_of(self, source: NodeId, dest: NodeId) -> DiGraph:
        """
        Получить связный подграф между двумя заданными узлами графа InputOneOf
        """

        return get_connected_subgraph(nx.subgraph_view(self.dag.graph), source, dest)

    async def _add_case_result(self, ctx: PipelineContextLike, switch_node_id: NodeId) -> None:
        """
        Записать выбор, сделанный оператором выбора Switch-Case, в контекст
        """
        selected_branch_label = None
        branch_nodes = {}
        for pred_id in self.dag.graph.predecessors(switch_node_id):
            edge = self.dag.graph.edges[(pred_id, switch_node_id)]

            if edge.get(EdgeField.is_switch):
                selected_branch_label = await ctx.load_node_result(pred_id)
                continue

            branch_nodes[edge.get(EdgeField.case_branch)] = pred_id

        await ctx.add_case_result(
            switch_node_id,
            CaseResult(label=selected_branch_label, node_id=branch_nodes[selected_branch_label]),
        )

    async def _run_node(
        self,
        ctx: PipelineContextLike,
        node_id: NodeId,
        is_node_from_success_pool: bool,
        force_default: bool = False,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Выполнить узел графа
        """

        await ctx.add_node_in_run(node_id)
        await ctx.emit_on_node_start(node_id=node_id)

        try:
            logger.info('Начало исполнения ноды, node_id=%s', node_id)
            input_kwargs = await self._get_node_kwargs(ctx, node_id)

            # Если в узел попадают параметры, которые являются ошибками, то исполнять этот узел нельзя.
            # Как следствие, нам нужно завершить исполнение текущей ноды
            for key_value in input_kwargs.values():
                if isinstance(key_value, Exception):
                    raise key_value

            result = await self.execute(
                ctx=ctx,
                node_id=node_id,
                force_default=force_default,
                **input_kwargs,
            )

            if isinstance(result, Exception):
                await ctx.emit_on_node_complete(node_id=node_id, error=result)
            else:
                await ctx.emit_on_node_complete(node_id=node_id, error=None)

            logger.info('Завершение исполнения ноды, node_id=%s', node_id)
            return result

        except Exception as ex:
            await ctx.emit_on_node_complete(node_id=node_id, error=ex)
            logger.error('Ошибка исполнения ноды node_id=%s, err=%s', node_id, ex)

            if is_node_from_success_pool:
                return NodeErrorType.succession_node_error

            raise ex

    async def execute(
        self,
        ctx: PipelineContextLike,
        node_id: NodeId,
        force_default: bool,
        **kwargs,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Запуск ноды с дополнительной обработкой

        Args:
            node_id: Идентификатор ноды
            ctx: Контекст исполнения узла
            force_default: Отдать дефолтное значение
            **kwargs: Ключевые аргументы целевой функции
        """

        serialized_node: NodeSerializerLike = self.dag.node_map[node_id]
        node = serialized_node.get_node()

        retry_policy = DagRetryPolicy(node=node)

        n_attempts = 1
        while True:
            try:
                logger.debug('Начало исполнения узла node_id=%s', node_id)
                if force_default:
                    return run_node_default(node, **kwargs)

                return await run_node(**kwargs, node=node)

            except retry_policy.exceptions as error:
                logger.debug(
                    'Node %s will be restarted in %s seconds...',
                    node_id,
                    retry_policy.delay,
                    exc_info=error,
                )

                if n_attempts == retry_policy.attempts:

                    if retry_policy.use_default:
                        return run_node_default(node, **kwargs)

                    raise error

                await ctx.emit_on_node_complete(node_id=node_id, error=error)

                n_attempts += 1
                await asyncio.sleep(retry_policy.delay)

            except Exception:
                if retry_policy.use_default:
                    return run_node_default(node, **kwargs)
                raise

    @staticmethod
    async def _get_call_order(ctx: PipelineContextLike, dag: DiGraph) -> t.List[NodeId]:
        """
        Рассчитываем план выполнения узлов графа.
        Учитываем, что если подграф находится в активной рекурсии, то исполнять нужно все узлы
        """

        return [
            node_id for node_id in nx.topological_sort(dag)
            if (not await ctx.is_node_in_run(node_id) if not dag.is_recurrent else True)
        ]

    def _get_node_dependencies(self, dag: DiGraph, node_id: NodeId) -> t.Set[NodeId]:
        """
        Метод возвращает узлы без которых оператор switch-case, oneof не запустится в указанном dag.
        """

        node_predecessors = set(self.dag.graph.predecessors(node_id))
        current_dag = set(nx.topological_sort(dag))

        return current_dag.intersection(node_predecessors)

    async def _get_predecessors(self, dag: DiGraph, ctx: PipelineContextLike, node_id: NodeId) -> t.List[NodeId]:
        """
        Получение всех предшественников для узла
        """
        predecessors = list(
            self._get_node_dependencies(dag, node_id)
            if self._is_switch(dag, node_id) or self._is_head_of_oneof(dag, node_id) or dag.is_recurrent
            else self.dag.graph.predecessors(node_id)
        )

        for idx, node_id in enumerate(predecessors):

            if self._is_switch(dag, node_id):
                with suppress(KeyError):
                    predecessors[idx] = (await ctx.get_case_result(node_id)).node_id

        return predecessors

    async def _is_ready_to_run(self, dag: DiGraph, ctx: PipelineContextLike, node_id: NodeId) -> bool:
        """
        Метод проверяет возможность запуска узла
        """

        if await ctx.is_node_in_run(node_id):
            return False

        for pred_node_id in await self._get_predecessors(dag, ctx, node_id):

            if not await ctx.exists_node_result(pred_node_id):
                return False

        return True

    async def _run_dag(
        self,
        ctx: PipelineContextLike,
        dag: DiGraph,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Запустить граф / подграф
        """

        logger.debug('Начало запуска DAG, dag=%s', str(dag))

        list_node_ids = await self._get_call_order(ctx, dag)

        if dag.is_recurrent:
            ctx.delete_node_results(list_node_ids)

        if len(list_node_ids) == 0:
            return

        dag_output_node = list_node_ids[-1]
        list_node_ids = deque(list_node_ids)

        awaitable_nodes = []
        awaitable_nodes_ids = []

        while list_node_ids:

            node_id = list_node_ids.popleft()
            logger.debug('Обработка узла node_id=%s', node_id)

            if (
                await self._is_ready_to_run(dag, ctx, node_id)
                and node_id not in awaitable_nodes_ids
            ):
                logger.debug('Узел готов к обработке node_id=%s', node_id)

                # Как только доходим до оператора Switch, находим подграф выбранной ветки Case
                # и рекурсивно выполняем его
                if self._is_switch(dag, node_id):
                    logger.debug('Узел является switch-кейсом node_id=%s', node_id)
                    await self._add_case_result(ctx, node_id)
                    await self._run_dag(
                        ctx=ctx,
                        dag=self._get_reduced_dag(
                            self.dag.input_node,
                            (await ctx.get_case_result(switch_node_id=node_id)).node_id,
                        ),
                    )

                # Запускаем рекурсию, если нода является родителем InputOneOf пула
                elif dag.nodes[node_id].get(NodeField.is_first_success):
                    logger.debug('Узел является InputOneOf node_id=%s', node_id)
                    await self._run_dag(
                        ctx=ctx,
                        dag=self._get_reduced_dag_input_one_of(
                            self.dag.input_node,
                            node_id,
                        ),
                    )

                else:
                    logger.debug('Узел может быть обработан, отправляем его в обработку node_id=%s', node_id)

                    awaitable_nodes_ids.append(node_id)
                    awaitable_nodes.append(
                        self._run_node(
                            ctx,
                            node_id,
                            # Добавляем флаг, что выполняется нода из InputOneOf пула
                            is_node_from_success_pool=self._is_node_in_oneof(dag, node_id),
                        )
                    )

                # Если мы только запустили последний узел, то результат необходимо получить из соседнего процесса,
                # по этой причине перезапускаем узел для повторной обработки
                if node_id == dag_output_node:
                    list_node_ids.appendleft(node_id)

            elif await ctx.is_node_in_run(node_id):

                with suppress(ValueError, IndexError):
                    node_id_index: int = awaitable_nodes_ids.index(node_id)
                    del awaitable_nodes[node_id_index]
                    del awaitable_nodes_ids[node_id_index]

            else:
                # В случае, если существуют не все зависимости, то ожидаем результатов исполнения зависимых узлов.
                # Топологическая сортировка дает гарантию, что здесь мы будем всегда ожидать необходимое количество
                # зависимостей для исполнения текущего узла, который после получения зависимостей будет отправлен
                # на повторную обработку
                nodes_result = await asyncio.gather(*awaitable_nodes)

                for idx, node_result in enumerate(nodes_result):
                    node_result_id = awaitable_nodes_ids[idx]

                    if isinstance(node_result, Recurrent):
                        max_iterations = dag.nodes[node_result_id].get(NodeField.max_iterations)
                        start_from_node_id = dag.nodes[node_result_id].get(NodeField.start_node)

                        if not await ctx.is_active_recurrence_subgraph(start_from_node_id, node_result_id):
                            await ctx.set_active_recurrence_subgraph(start_from_node_id, node_result_id)

                            recurrent_subgraph = get_connected_subgraph(
                                self.dag.graph, start_from_node_id, node_result_id, is_recurrent=True,
                            )

                            for current_iter in range(max_iterations):
                                self.dag.graph.nodes[start_from_node_id][NodeField.additional_data] = node_result.data

                                node_result = await self._run_dag(dag=recurrent_subgraph, ctx=ctx)

                                if current_iter + 1 == max_iterations and isinstance(node_result, Recurrent):
                                    node_result = await self._run_node(
                                        ctx,
                                        node_result_id,
                                        self._is_node_in_oneof(dag, node_result_id),
                                        force_default=True,
                                    )

                                elif not isinstance(node_result, Recurrent):
                                    break

                            await ctx.remove_recurrence_subgraph(start_from_node_id, node_result_id)

                        else:
                            # В случае, если исполняемый узел рекуррентного подграфа не завершается ожидаемым образом,
                            # то его нужно вернуть обратно к управляющей конструкции
                            return node_result

                    # Если нода из списка InputOneOf выполнилась успешно, то проходить остальные не нужно.
                    # При этом, результат сохраняется под выходным узлом
                    if (
                        not isinstance(node_result, NodeErrorType)
                        and self._is_node_in_oneof(dag, node_result_id)
                    ):
                        await ctx.save_node_result(dag_output_node, node_result)
                        # Для искусственной остановки пайплайна помечаем текущий узел как выходной
                        node_result_id = dag_output_node

                    else:
                        # Если узел обычный, то сохраняем результат в обычном режиме под оригинальным node_id
                        await ctx.save_node_result(node_result_id, node_result)

                    logger.debug(
                        'Получен результат узла node_id=%s, от которого зависит исполнение узла node_id=%s',
                        node_result_id,
                        node_id,
                    )

                    if node_result_id == dag_output_node:
                        return await ctx.load_node_result(node_result_id)

                awaitable_nodes.clear()
                awaitable_nodes_ids.clear()
                list_node_ids.appendleft(node_id)

    def __repr__(self):
        return f'<{self.__class__.__name__} nnodes="{len(self.dag.graph.nodes)}" nedges="{len(self.dag.graph.edges)}">'



from ml_pipeline_engine.node.enums import *  # noqa
from ml_pipeline_engine.node.errors import *  # noqa
from ml_pipeline_engine.node.node import *  # noqa
from ml_pipeline_engine.node.serialization import *  # noqa



import asyncio
import functools
import inspect
import typing as t
import uuid

from ulid import ULID

from ml_pipeline_engine.module_loading import get_instance
from ml_pipeline_engine.node.enums import NodeTag
from ml_pipeline_engine.node.errors import (
    ClassExpectedError,
    RunMethodExpectedError,
)
from ml_pipeline_engine.parallelism import (
    process_pool_registry,
    threads_pool_registry,
)
from ml_pipeline_engine.types import NodeBase, NodeId, NodeLike

NodeResultT = t.TypeVar('NodeResultT')


def generate_pipeline_id() -> uuid.UUID:
    return ULID().to_uuid()


def generate_node_id(prefix: str, name: t.Optional[str] = None) -> str:
    return f'{prefix}__{name if name is not None else uuid.uuid4().hex[-8:]}'


def get_node_id(node: NodeLike) -> NodeId:
    node_type = node.node_type if getattr(node, 'node_type', None) else 'node'  # noqa

    if getattr(node, 'name', None):
        node_name = node.name  # noqa
    else:
        node_name = f'{node.__module__}_{getattr(node, "__name__", node.__class__.__name__)}'.replace('.', '_')

    return '__'.join([node_type, node_name])


def get_run_method(node: NodeLike) -> t.Optional[str]:
    run_method = None

    for method in NodeBase.RUN_METHOD_ALIASES:
        if callable(getattr(node, method, None)):
            if run_method is not None:
                raise AssertionError(f'Node should have only one run method. {run_method} + {method} detected')
            run_method = method

    return run_method


def get_callable_run_method(node: NodeLike) -> t.Callable:
    run_method_name = get_run_method(node)

    if run_method_name is not None:
        node = get_instance(node)
        return getattr(node, run_method_name)

    return node


def run_node_default(node: NodeLike[NodeResultT], **kwargs) -> t.Type[NodeResultT]:
    """
    Запуск получения дефолтного значения узла
    """
    return get_instance(node).get_default(**kwargs)


async def run_node(node: NodeLike[NodeResultT], *args, **kwargs) -> t.Type[NodeResultT]:
    """
    Функция для запуска узла.
    Запуск учитывает наличие тега для декларирования запуска узлов.

    Правила запуска узла:
        1. Если узел корутина, то запускается в текущем потоке.
           Ожидается, что если мы запускаем корутину, то под капотом не будет сложных вычислений.
        2. Если узел не корутина, то смотрим в теги и выбираем между процессом или потоком
        3. Если узел не корутина и тегов нет, то исполняем в потоке
    """

    run_method = get_callable_run_method(node)
    loop = asyncio.get_running_loop()
    tags = node.tags or ()

    if inspect.iscoroutinefunction(run_method):
        result = await run_method(*args, **kwargs)

    elif NodeTag.non_async in tags:
        result = run_method(*args, **kwargs)

    else:
        executor = (
            process_pool_registry.get_pool_executor()
            if NodeTag.process in tags
            else threads_pool_registry.get_pool_executor()
        )

        result = await loop.run_in_executor(
            executor,
            functools.partial(run_method, *args, **kwargs),
        )

    return result


def build_node(
    node: NodeLike,
    node_name: t.Optional[str] = None,
    class_name: t.Optional[str] = None,
    atts: t.Optional[t.Dict[str, t.Any]] = None,
    dependencies_default: t.Optional[t.Dict[str, t.Any]] = None,
    **target_dependencies,
) -> t.Type[NodeLike]:
    """
    Функция создает новый узел графа на основе generic-узлов.
    НЕ generic узел отличается тем, что целевой метод начинает зависеть от конкретных узлов

    Args:
        node: Класс ноды
        class_name: Название класса узла
        node_name: Название ноды для соблюдения протокола
        atts: Дополнительные атрибуты нового объекта
        dependencies_default: Дефолтные значения для зависимостей
        **target_dependencies: Целевые зависимости generic-зависимостей
    """

    if not inspect.isclass(node):
        raise ClassExpectedError('Для создания узла ожидается объекта класса')

    run_method = get_run_method(node)
    if not run_method:
        raise RunMethodExpectedError(
            f'Ожидается наличие хотя бы одного run-метода. methods={NodeBase.RUN_METHOD_ALIASES}',
        )

    if inspect.iscoroutinefunction(getattr(node, run_method)):

        async def class_method(*args, **kwargs):
            return await getattr(node, run_method)(*args, **kwargs, **(dependencies_default or {}))

    else:
        def class_method(*args, **kwargs):
            return getattr(node, run_method)(*args, **kwargs, **(dependencies_default or {}))

    class_name = class_name or f'Generic{node.__name__}'
    created_node = type(
        class_name,
        (node,),
        {
            # Меняем на lambda-функцию, чтобы убить ссылку на метод родительского класса.
            run_method: class_method,
            '__module__': __name__,
            'name': node_name or node.name,  # noqa
            **(atts or {}),
        },
    )

    method = getattr(created_node, run_method)
    method.__annotations__.update(target_dependencies)

    globals()[class_name] = created_node
    return created_node



import abc
import typing as t
from dataclasses import dataclass
from uuid import UUID

import networkx as nx

ProcessorResultT = t.TypeVar('ProcessorResultT')
DataSourceResultT = t.TypeVar('DataSourceResultT')
FeatureResultT = t.TypeVar('FeatureResultT')
FeatureVectorizerResultT = t.TypeVar('FeatureVectorizerResultT')
MLModelResultT = t.TypeVar('MLModelResultT')

NodeResultT = t.TypeVar('NodeResultT')
AdditionalDataT = t.TypeVar('AdditionalDataT', bound=t.Any)

PipelineId = t.Union[UUID, str]

NodeId = str
ModelName = str
CaseLabel = str
SerializationNodeKind = str
NodeTag = str


class RetryProtocol(t.Protocol):
    """
    Протокол, описывающий свойства повторного исполнения ноды.
    Предоставляет возможность использовать дефолтное значение в случае ошибки.
    """

    attempts: t.ClassVar[t.Optional[int]] = None
    delay: t.ClassVar[t.Optional[t.Union[int, float]]] = None
    exceptions: t.ClassVar[t.Optional[t.Tuple[t.Type[BaseException], ...]]] = None
    use_default: t.ClassVar[bool] = False

    def get_default(self, **kwargs) -> NodeResultT:
        ...


class TagProtocol(t.Protocol):
    """
    Протокол, который позволяет декларировать параметры запуска узла
    """

    tags: t.ClassVar[t.Tuple[NodeTag, ...]] = ()


@dataclass
class Recurrent:
    data: t.Optional[AdditionalDataT]


class RecurrentProtocol(t.Protocol):
    """
    Протокол, определяющий поведение узла при повторном исполнении подграфа
    """

    @abc.abstractmethod
    def next_iteration(self, data: t.Optional[AdditionalDataT]) -> Recurrent:
        """
        Метод для инициализации повторного исполнения подграфа
        """


class NodeProtocol(t.Protocol):
    """
    Узел графа модели
    """

    RUN_METHOD_ALIASES = (
        'process',
        'extract',
        'collect',
        'vectorize',
        'predict',
    )

    node_type: t.ClassVar[str] = None
    name: t.ClassVar[str] = None
    title: t.ClassVar[str] = None


class NodeBase(NodeProtocol, RetryProtocol, TagProtocol):
    pass


class ProcessorLike(RetryProtocol, TagProtocol, t.Protocol[ProcessorResultT]):  # noqa
    """
    Узел общего назначения
    """

    process: t.Union[
        t.Callable[..., ProcessorResultT],
        t.Callable[..., t.Awaitable[ProcessorResultT]]
    ]


class DataSourceLike(RetryProtocol, TagProtocol, t.Protocol[DataSourceResultT]):  # noqa
    """
    Источник данных
    """

    collect: t.Union[
        t.Callable[..., DataSourceResultT],
        t.Callable[..., t.Awaitable[DataSourceResultT]]
    ]


class FeatureLike(RetryProtocol, TagProtocol, t.Protocol[FeatureResultT]):  # noqa
    """
    Фича
    """

    extract: t.Union[
        t.Callable[..., FeatureResultT],
        t.Callable[..., t.Awaitable[FeatureResultT]]
    ]


class FeatureVectorizerLike(RetryProtocol, TagProtocol, t.Protocol[FeatureVectorizerResultT]):
    """
    Векторизатор фичей
    """

    vectorize: t.Union[
        t.Callable[..., FeatureVectorizerResultT],
        t.Callable[..., t.Awaitable[FeatureVectorizerResultT]]
    ]


class MLModelLike(RetryProtocol, TagProtocol, t.Protocol[MLModelResultT]):
    """
    ML-модель
    """

    predict: t.Union[
        t.Callable[..., MLModelResultT],
        t.Callable[..., t.Awaitable[MLModelResultT]]
    ]


NodeLike = t.Union[
    t.Type[ProcessorLike[NodeResultT]],
    t.Type[DataSourceLike[NodeResultT]],
    t.Type[FeatureLike[NodeResultT]],
    t.Type[FeatureVectorizerLike[NodeResultT]],
    t.Type[MLModelLike[NodeResultT]],
]


class NodeSerializerLike(t.Protocol):
    node: t.Union[bytes]
    serialization_kind: SerializationNodeKind

    @classmethod
    def serialize(
        cls,
        node: NodeLike,
        serialization_kind: t.Optional[SerializationNodeKind] = None,
    ) -> 'NodeSerializerLike':
        ...

    def get_node(self) -> NodeLike:
        ...


@dataclass(frozen=True)
class PipelineResult(t.Generic[NodeResultT]):
    """
    Контейнер результата выполнения пайплайна.
    """

    pipeline_id: PipelineId
    value: NodeResultT
    error: t.Optional[Exception]

    def raise_on_error(self) -> None:
        if self.error is not None:
            raise self.error


@dataclass(frozen=True)
class CaseResult:
    """
    Результат оператора switch-case
    """

    label: CaseLabel
    node_id: NodeId


class PipelineChartLike(t.Protocol[NodeResultT]):
    """
    Определение пайплайна ML-модели

    Этот объект можно хранить в глобальной области. Он является иммутабельным.
    """

    model_name: ModelName
    entrypoint: t.Optional[t.Union[NodeLike[NodeResultT], 'DAGLike[NodeResultT]']]
    event_managers: t.List[t.Type['EventManagerLike']]
    artifact_store: t.Optional[t.Type['ArtifactStoreLike']]

    async def run(
        self,
        pipeline_id: t.Optional[PipelineId] = None,
        input_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        meta: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> PipelineResult[NodeResultT]:
        ...


class PipelineContextLike(t.Protocol):
    """
    Контекст выполнения пайплайна

    Создается при запуске пайплайна из чарта пайплайна и хранит состояние выполнения пайплайна, является источником
    событий жизненного цикла пайплайна.
    """

    chart: PipelineChartLike
    pipeline_id: PipelineId
    input_kwargs: t.Dict[str, t.Any]
    meta: t.Dict[str, t.Any]
    artifact_store: 'ArtifactStoreLike'

    async def emit_on_node_start(self, node_id: NodeId):
        ...

    async def emit_on_node_complete(self, node_id: NodeId, error: t.Optional[Exception]):
        ...

    async def emit_on_pipeline_start(self):
        ...

    async def emit_on_pipeline_complete(self, result: PipelineResult):
        ...

    async def save_node_result(self, node_id: NodeId, data: t.Any) -> None:
        ...

    async def load_node_result(self, node_id: NodeId) -> t.Any:
        ...

    async def add_case_result(self, switch_node_id: NodeId, selection: CaseResult) -> None:
        ...

    async def get_case_result(self, switch_node_id: NodeId) -> CaseResult:
        ...

    async def add_node_in_run(self, node_id: NodeId) -> None:
        ...

    async def is_node_in_run(self, node_id: NodeId) -> bool:
        ...

    async def exists_node_result(self, node_id: NodeId) -> bool:
        ...

    async def is_active_recurrence_subgraph(self, source: NodeId, dest: NodeId) -> bool:
        ...

    async def set_active_recurrence_subgraph(self, source: NodeId, dest: NodeId) -> None:
        ...

    async def remove_recurrence_subgraph(self, source: NodeId, dest: NodeId) -> None:
        ...

    def delete_node_results(self, node_ids: t.Iterable[NodeId]) -> None:
        ...

    @property
    def model_name(self) -> ModelName:  # noqa
        ...

    def _get_event_managers(self) -> t.List['EventManagerLike']:
        ...


class EventManagerLike(t.Protocol):
    """
    Менеджер событий жизненного цикла пайплайна
    """

    async def on_pipeline_start(self, ctx: PipelineContextLike) -> None:
        ...

    async def on_pipeline_complete(self, ctx: PipelineContextLike, result: PipelineResult) -> None:
        ...

    async def on_node_start(self, ctx: PipelineContextLike, node_id: NodeId) -> None:
        ...

    async def on_node_complete(self, ctx: PipelineContextLike, node_id: NodeId, error: t.Optional[Exception]) -> None:
        ...


class ArtifactStoreLike(t.Protocol):
    """
    Хранилище артефактов - результатов расчета узлов
    """

    def __init__(self, ctx: PipelineContextLike, *args, **kwargs):
        ...

    async def save(self, node_id: NodeId, data: t.Any) -> None:
        ...

    async def load(self, node_id: NodeId) -> NodeResultT:
        ...


class DAGCacheManagerLike(t.Protocol):
    """
    Менеджер локального кэша DAG
    """

    @abc.abstractmethod
    def save(self, node_id: NodeId, data: t.Any):
        ...

    @abc.abstractmethod
    def load(self, node_id: NodeId) -> t.Any:
        ...

    @abc.abstractmethod
    def exists(self, node_id: NodeId) -> bool:
        ...


class RetryPolicyLike(t.Protocol):
    node: NodeLike

    @property
    @abc.abstractmethod
    def delay(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def attempts(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def exceptions(self) -> t.Tuple[t.Type[Exception]]:
        ...


class DAGRunManagerLike(t.Protocol):
    """
    Менеджер управлением запуска графа
    """

    dag: 'DAGLike'

    @abc.abstractmethod
    async def run(self, ctx: PipelineContextLike) -> NodeResultT:
        ...


class DAGLike(t.Protocol[NodeResultT]):
    """
    Граф
    """

    graph: [t.Type[nx.DiGraph], nx.DiGraph]
    input_node: NodeId
    output_node: NodeId
    node_map: t.Dict[NodeId, NodeSerializerLike]
    run_manager: DAGRunManagerLike
    retry_policy: RetryPolicyLike
    is_process_pool_needed: bool
    is_thread_pool_needed: bool

    @abc.abstractmethod
    async def run(self, ctx: PipelineContextLike) -> NodeResultT:
        ...



from concurrent.futures import ThreadPoolExecutor

from ml_pipeline_engine.parallelism.basic import (
    PoolExecutorRegistry as BasePoolExecutorRegistry,
)

__all__ = ('threads_pool_registry',)


class PoolExecutorRegistry(BasePoolExecutorRegistry):

    def is_ready(self) -> None:
        if not self._pool_executor or self._pool_executor._shutdown:  # noqa
            raise RuntimeError('Исполнение невозможно без указания пула потоков')

    def auto_init(self):
        self.register_pool_executor(ThreadPoolExecutor())


threads_pool_registry = PoolExecutorRegistry()



import io
import json
import pickle
import typing as t
from abc import ABC, abstractmethod

from ml_pipeline_engine.artifact_store.enums import DataFormat

SerializableObjectT = t.Any


class SerializerInitializationError(Exception):
    pass


class Serializer(ABC):
    @abstractmethod
    def dump(self, obj: SerializableObjectT, fp: t.IO) -> None:
        ...

    @abstractmethod
    def load(self, fp: t.IO) -> SerializableObjectT:
        ...

    @abstractmethod
    def get_default_io(self) -> t.IO:
        ...


class PickleSerializer(Serializer):
    def dump(self, obj: SerializableObjectT, fp: t.IO) -> None:
        pickle.dump(obj, fp)
        fp.seek(0)

    def load(self, fp: t.IO) -> SerializableObjectT:
        fp.seek(0)
        return pickle.load(fp)

    def get_default_io(self) -> t.IO:
        return io.BytesIO()


class JSONSerializer(Serializer):
    def dump(self, obj: SerializableObjectT, fp: t.IO) -> None:
        json.dump(obj, fp, indent=4, ensure_ascii=False)
        fp.seek(0)

    def load(self, fp: t.IO) -> SerializableObjectT:
        fp.seek(0)
        return json.load(fp)

    def get_default_io(self) -> t.IO:
        return io.StringIO()


class SerializerFactory:
    @staticmethod
    def from_data_format(fmt: DataFormat) -> Serializer:
        if fmt == DataFormat.PICKLE:
            return PickleSerializer()

        return JSONSerializer()

    def from_extension(self, extension: str) -> Serializer:
        try:
            fmt = DataFormat(extension)
        except ValueError:
            raise SerializerInitializationError(f'No suitable serializer for {extension} extension')

        return self.from_data_format(fmt)


serializer_factory = SerializerFactory()



from dataclasses import dataclass
from typing import Callable, Optional, Union

import cloudpickle

from ml_pipeline_engine.node.enums import SerializationNodeKind
from ml_pipeline_engine.types import NodeLike, NodeSerializerLike

_SERIALIZATION_MAPPING = {
    SerializationNodeKind.cloudpickle: cloudpickle,
}


@dataclass
class NodeSerializer(NodeSerializerLike):
    node: Union[NodeLike, bytes]
    serialization_kind: SerializationNodeKind

    @classmethod
    def serialize(cls, node: NodeLike, serialization_kind: Optional[SerializationNodeKind] = None) -> 'NodeSerializer':
        """
        Автоматическая сериализация объекта узла в бинарный вид для возможности передачи между процессами и тредами

        :param node: Объект узла
        :param serialization_kind: Модуль для сериализации
        """
        if serialization_kind is None:
            serialization_kind = SerializationNodeKind.cloudpickle

        return cls(
            node=cls._dumps(serialization_kind)(node),
            serialization_kind=serialization_kind,
        )

    def get_node(self) -> NodeLike:
        """
        Получение десериализованной ноды
        """
        return self._loads(self.serialization_kind)(self.node)

    @staticmethod
    def _dumps(serialization_kind: SerializationNodeKind) -> Callable:
        return _SERIALIZATION_MAPPING[serialization_kind].dumps

    @staticmethod
    def _loads(serialization_kind: SerializationNodeKind) -> Callable:
        return _SERIALIZATION_MAPPING[serialization_kind].loads



from ml_pipeline_engine.types import (
    AdditionalDataT,
    NodeBase,
    Recurrent,
    RecurrentProtocol,
)


class ProcessorBase(NodeBase):
    """
    Базовый класс для обработчиков общего назначения
    """

    node_type = 'processor'

    def process(self, *args, **kwargs):
        raise NotImplementedError('Method process() is not implemented')


class RecurrentProcessor(ProcessorBase, RecurrentProtocol):
    """
    Узел процессора, который может быть исполнен в рекуррентном подграфе
    """

    def next_iteration(self, data: AdditionalDataT) -> Recurrent:
        return Recurrent(data=data)



from ml_pipeline_engine.dag.dag import *  # noqa
from ml_pipeline_engine.dag.enums import *  # noqa
from ml_pipeline_engine.dag.manager import *  # noqa
from ml_pipeline_engine.dag.retrying import *  # noqa



import sys
import typing as t
from importlib import import_module

__all__ = [
    'import_string',
    'get_instance',
]


def _cached_import(module_path, class_name) -> t.Type:
    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, '__spec__', None))
        and getattr(spec, '_initializing', False) is False  # noqa
    ):
        module = import_module(module_path)
    return getattr(module, class_name)


def import_string(dotted_path) -> t.Type:
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    try:
        return _cached_import(module_path, class_name)
    except AttributeError as err:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class') from err


def get_instance(cls: t.Type, *args, **kwargs) -> t.Any:
    default_factory = getattr(cls, 'default_factory', None)
    if default_factory is None:
        return cls(*args, **kwargs)
    return default_factory(*args, **kwargs)
